/*
	PCPORTS.C -- Serial Support for PC Compatibles (8250 UART)

	PC89LINK: HDOS <-> MSDOS File Transfer Package

	Copyright (c) 1989 Lindley Systems
			   4257 Berwick Place
			   Woodbridge, VA 22192


*/

#include <conio.h>
extern int portnum;

/* Prototypes: */

unsigned char nget();
void setup ();				/* set up I/O ports */
void init(int entry);
int nports();				/* Number of defined ports */
void nput (char c);			/* send	c to port */
int tget();				/* Get a character, with time-out */
int tstport();				/* test	input port */
void initserial();			/* Set up interrupts &c. */
void resetserial();

#define CHARTIME -1		/* Timeout delay for each character */

#define	SERPORTS 0x00400000L	/* BIOS port table */
#define MAXPORT 4		/* COM4: is last number */

#define PORT_CHAR  	0	/* Transmit/Receive buffer (DLAB=0) */
#define PORT_INT_ENAB	1	/* Interrupt Enable register (DLAB=0) */
#define PORT_LATCH_LSB 	0	/* Divisor latch LSB (DLAB=1) */
#define PORT_LATCH_MSB	1	/* Divisor latch MSB (DLAB=1) */
#define PORT_INT_ID	2	/* Interrupt Identification */
#define PORT_LCR	3	/* Line control register */
#define PORT_MCR	4	/* Modem control register */
#define PORT_LSR	5	/* Line status register */
#define PORT_MSR	6	/* Modem status register */

#define INT_EN_NONE	0	/* no interrupts */
#define INT_EN_DATA	1	/* Enable Data Available interrupt */
#define INT_EN_THRE	2	/* Enable Transmit Hold Reg Empty interrupt */
#define INT_EN_RLS	4	/* Enable Receive Line Status interrupt */
#define INT_EN_MS	8	/* Enable Modem Status interrupt */

/* Line Control */

#define LCR_5BITS	0	/* 5-bit words */
#define LCR_6BITS	1	/* 6-bit words */
#define LCR_7BITS	2	/* 7-bit words */
#define LCR_8BITS	3	/* 8-bit words */
#define LCR_1STOP	0	/* 1 Stop bit */
#define LCR_2STOP	4	/* 2 Stop bits */
#define LCR_NOPAR	0	/* No parity */
#define LCR_EVEN_PAR	8+16	/* Even Parity */
#define LCR_ODD_PAR	8	/* Odd Parity */
#define LCR_BREAK	64	/* Send break */
#define LCR_DLAB	128	/* Divisor latch access */

/* Line Status */

#define LSR_READY	1	/* Data ready */
#define LSR_E_OVERRUN	2	/* Overrun error */
#define LSR_E_PARITY	4	/* Parity error */
#define LSR_E_FRAMING	8	/* Framing error */
#define LSR_I_BREAK	16	/* Break interrupt */
#define LSR_THRE	32	/* Transmitter holding register empty */
#define LSR_TSRE	64	/* Transmitter shift register empty */

#define LSR_ERROR	LSR_E_OVERRUN + LSR_E_PARITY + LSR_E_FRAMING

/* Modem Control */

#define MCR_DTR		1
#define MCR_RTS		2
#define MCR_OUT1 	4	/* unused on PC */
#define MCR_OUT2 	8	/* Set to enable interrupts */
#define MCR_LOOP	16	/* Set for loopback operation */

/* Modem Status */

#define MSR_DCTS	1	/* Change in CTS */
#define MSR_DDSR	2	/* Change in DSR */
#define MSR_TERI	4	/* Trailing edge ring indicator */
#define MSR_DRLSD	8	/* Change Rx Line signal detect */
#define MSR_CTS		16	/* CTS */
#define MSR_DSR		32	/* DSR */
#define MSR_RI		64	/* Ring indicator */
#define MSR_RLSD	128	/* Receive line signal detect */

int port = 0;			/* I/O base port */

int maxbaud = 9;		/* Entries in table */
int baudent = 3;		/* Default to this entry in baudrate table */
long baudtable[9]=
	{1200,2400,4800,9600,12800,19200,38400,57600,115200};
static long baud_div [9]=
	{  96,  48,  24,  12,    9,    6,    3,    2,     1};

/* Port numbers to use if BIOS doesn't have them */
static int default_port[4]={0x3f8,0x2f8,0x3e8,0x2e8};

static int valid_port;			/* TRUE if init() finds UART at */
					/* selected port # */

void get_port_addr()
{
  port = *((int far *)SERPORTS+portnum);
  if (port == 0)			/* Port not supported by BIOS? */
    port = default_port[portnum];	/* Assume standard I/O address */
}

void setup ()				/* set up I/O ports */
{
  get_port_addr();
  init(baudent);
}

void loopback(int type)
{
  switch (type) {
    case 0:
	outp(port+PORT_MCR, MCR_DTR + MCR_RTS); /* clear loop-back mode */
	break;
    case 1:
	outp(port+PORT_MCR, MCR_LOOP); /* set loop-back mode */
	break;
  }
}

/* Find number of serial ports */

int nports()
{ int count = 0, old_portnum;

	old_portnum = portnum;		/* Save configured port # */
	for (portnum=0; portnum < MAXPORT; portnum++)
	  { if (port_exists(portnum))	/* See if port's really there */
	      count++;
	  }
	portnum = old_portnum;		/* Go back to configured port */
	return(count);			/* and return # of ports */
}

void initserial()
{}

void resetserial()
{}

int port_exists()
{
  int lcr_val, its_there;

  get_port_addr();			/* find port's i/o addr */
  lcr_val = inp(port+PORT_LCR);		/* determine whether reading divisor */
  outp(port+PORT_LCR,lcr_val ^ LCR_DLAB); /* flip DLAB access bit */
  its_there = inp(port+PORT_LCR) != lcr_val; /* did it change? */
  outp(port+PORT_LCR,lcr_val);		/* restore previous setting */
  return(its_there);
}

void init (int entry)			/* initialize port */
{
  outp(port+PORT_INT_ENAB, INT_EN_NONE); /* clear interrupts */
  outp(port+PORT_MCR, MCR_LOOP);	/* set loop-back mode */
  outp(port+PORT_LCR, LCR_DLAB);	/* set divisor latch */
  outp(port+PORT_LATCH_MSB, 0);	/* Data rates faster than 300 */
  outp(port+PORT_LATCH_LSB, baud_div[entry]);	/* Divisor for this rate */
  valid_port = (inp(port) == baud_div[entry]);
					/* If we can read it back, valid port. */

  outp(port+PORT_LCR, LCR_8BITS + LCR_1STOP + LCR_NOPAR);
					/* word	length = 8 bits	*/
  inp(port);				/* read	a char */
  wait (16);				/* wait	two char times (or more) */
  inp(port);				/* read	another	char */
  outp(port+PORT_MCR, MCR_DTR + MCR_RTS); /* clear loop-back mode */
}

void nput (char c)			/* send	c to port */
{
  while	((inp(port + 5)	& 32) == 0);	/* wait	till transmitter ready */
  outp(port, c);			/* send	the char */
}

/* TGET - Timeout Get character. Return character, or -1 if none */

static unsigned char status;		/* last line status */

int tstport()				/* test	input port */
{
  status = inp(port + PORT_LSR);
  /* return 0=no data, 1=data, -1=error */
  return((status & LSR_READY)==0?0:(status&(LSR_E_OVERRUN|LSR_E_FRAMING))?-1:1);
}

int tget()				/* Get a character, with time-out */
{ int t=CHARTIME;			/* Countdown delay */
	while (tstport() <= 0)
	  if (--t ==0)
	    return(-1);			/* No character */
	return((unsigned)inp(port));	/* Return char */
}

unsigned char nget ()			/* get a character */
{
  while	(tstport() <= 0);		/* wait	for a character	*/
  return (unsigned)inp(port);		/* read	char, and return it */
}

int rcv_pkt(int pktlen, char *buf)	/* Returns -1 if error */
{
  register char *p;
  register int count;
  int timeout;				/* If this counts down to 0, time out */

	count = pktlen;
	p = buf;
	timeout=CHARTIME;
	do
	  { if (inp(port+5)&1)		/* Code from TSTPORT() */
	      { *p++ = (unsigned)inp(port); /* Code from NGET() */
	        count--;
		timeout=CHARTIME;
	      }
	    else
	      timeout--;
	  }
	while (timeout!=0 && count>0);
	return (timeout==0);		/* Return -1 if counted down to 0 */
}
